@@ -3,7 +3,7 @@ from typing import Tuple, Union
 import numpy as np
 
 from highway_env.road.road import Road, Route, LaneIndex
-from highway_env.types import Vector
+from highway_env.utils import Vector
 from highway_env.vehicle.controller import ControlledVehicle
 from highway_env import utils
 from highway_env.vehicle.kinematics import Vehicle
@@ -143,7 +143,7 @@ class IDMVehicle(ControlledVehicle):
         """
         if not ego_vehicle or not isinstance(ego_vehicle, Vehicle):
             return 0
-        ego_target_speed = utils.not_zero(getattr(ego_vehicle, "target_speed", 0))
+        ego_target_speed = abs(utils.not_zero(getattr(ego_vehicle, "target_speed", 0)))
         acceleration = self.COMFORT_ACC_MAX * (
                 1 - np.power(max(ego_vehicle.speed, 0) / ego_target_speed, self.DELTA))
 
@@ -170,34 +170,6 @@ class IDMVehicle(ControlledVehicle):
         d_star = d0 + ego_vehicle.speed * tau + ego_vehicle.speed * dv / (2 * np.sqrt(ab))
         return d_star
 
-    def maximum_speed(self, front_vehicle: Vehicle = None) -> Tuple[float, float]:
-        """
-        Compute the maximum allowed speed to avoid Inevitable Collision States.
-
-        Assume the front vehicle is going to brake at full deceleration and that
-        it will be noticed after a given delay, and compute the maximum speed
-        which allows the ego-vehicle to brake enough to avoid the collision.
-
-        :param front_vehicle: the preceding vehicle
-        :return: the maximum allowed speed, and suggested acceleration
-        """
-        if not front_vehicle:
-            return self.target_speed
-        d0 = self.DISTANCE_WANTED
-        a0 = self.COMFORT_ACC_MIN
-        a1 = self.COMFORT_ACC_MIN
-        tau = self.TIME_WANTED
-        d = max(self.lane_distance_to(front_vehicle) - self.LENGTH / 2 - front_vehicle.LENGTH / 2 - d0, 0)
-        v1_0 = front_vehicle.speed
-        delta = 4 * (a0 * a1 * tau) ** 2 + 8 * a0 * (a1 ** 2) * d + 4 * a0 * a1 * v1_0 ** 2
-        v_max = -a0 * tau + np.sqrt(delta) / (2 * a1)
-
-        # Speed control
-        self.target_speed = min(self.maximum_speed(front_vehicle), self.target_speed)
-        acceleration = self.speed_control(self.target_speed)
-
-        return v_max, acceleration
-
     def change_lane_policy(self) -> None:
         """
         Decide when to change lane.
@@ -207,7 +179,7 @@ class IDMVehicle(ControlledVehicle):
         - closeness of the target lane;
         - MOBIL model.
         """
-        # If a lane change already ongoing
+        # If a lane change is already ongoing
         if self.lane_index != self.target_lane_index:
             # If we are on correct route but bad lane: abort it if someone else is already changing into the same lane
             if self.lane_index[:2] == self.target_lane_index[:2]:
@@ -233,6 +205,9 @@ class IDMVehicle(ControlledVehicle):
             # Is the candidate lane close enough?
             if not self.road.network.get_lane(lane_index).is_reachable_from(self.position):
                 continue
+            # Only change lane when the vehicle is moving
+            if np.abs(self.speed) < 1:
+                continue
             # Does the MOBIL model recommend a lane change?
             if self.mobil(lane_index):
                 self.target_lane_index = lane_index
@@ -258,7 +233,7 @@ class IDMVehicle(ControlledVehicle):
         # Do I have a planned route for a specific lane which is safe for me to access?
         old_preceding, old_following = self.road.neighbour_vehicles(self)
         self_pred_a = self.acceleration(ego_vehicle=self, front_vehicle=new_preceding)
-        if self.route and self.route[0][2]:
+        if self.route and self.route[0][2] is not None:
             # Wrong direction
             if np.sign(lane_index[2] - self.target_lane_index[2]) != np.sign(self.route[0][2] - self.target_lane_index[2]):
                 return False
